/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hortonworks.spark.sql.hive.llap;

import com.google.common.base.Preconditions;
import com.hortonworks.hwc.MergeBuilder;
import com.hortonworks.spark.sql.hive.llap.common.HWConf;
import com.hortonworks.spark.sql.hive.llap.common.HWConf.HWCReadMode;
import com.hortonworks.spark.sql.hive.llap.common.HwcResource;
import com.hortonworks.spark.sql.hive.llap.query.builder.CreateTableBuilder;
import com.hortonworks.spark.sql.hive.llap.common.DriverResultSet;
import com.hortonworks.spark.sql.hive.llap.util.*;
import com.hortonworks.spark.sql.hive.llap.common.HwcSparkListener;
import com.hortonworks.spark.sql.hive.llap.query.builder.MergeBuilderImpl;
import com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.hadoop.hive.llap.LlapBaseInputFormat;
import org.apache.spark.SparkConf;
import org.apache.spark.sql.DataFrameReader;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.RuntimeConfig;
import org.apache.spark.sql.SparkSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.qubole.spark.hiveacid.transaction.HiveAcidTxnManagerObject;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Supplier;
import java.util.regex.Pattern;

import org.apache.hadoop.fs.permission.FsPermission;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.conf.Configuration;


import static com.hortonworks.spark.sql.hive.llap.common.HWConf.DEFAULT_DB;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_HIVE_JDBC;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_QUERY_LLAP;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.isDataFetchQuery;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.resolveExecutionMethod;
import static com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataWriterHelper.FileFormat.ORC;
import static java.lang.String.format;

public class HiveWarehouseSessionImpl extends com.hortonworks.hwc.HiveWarehouseSession {

  private static final Logger LOG = LoggerFactory.getLogger(HiveWarehouseSessionImpl.class);

  private static final String SELECT_STMT_REGEX = "(\\s+)?select(\\s+).+";
  private static final Pattern selectPattern = Pattern.compile(SELECT_STMT_REGEX, Pattern.CASE_INSENSITIVE);
  public static final String SESSION_CLOSED_MSG = "Session is closed!!! No more operations can be performed. Please create a new HiveWarehouseSession";
  public static final String HWC_SESSION_ID_KEY = "hwc_session_id";
  public static final String HWC_QUERY_MODE = "hwc_query_mode";
  static String HIVE_WAREHOUSE_CONNECTOR_INTERNAL = HiveWarehouseSession.HIVE_WAREHOUSE_CONNECTOR;
  private static String STAGING_OUTPUT_DIR = "/tmp/SetStagingHwcDir";
  private static Boolean STAGING_OUTPUT = false;
  protected HiveWarehouseSessionState sessionState;

  protected Supplier<Connection> getConnector;

  protected Supplier<Connection> getConnOutsidePool;

  protected TriFunction<Connection, String, String, DriverResultSet> executeStmt;

  protected TriFunction<Connection, String, String, Boolean> executeUpdate;

  protected FunctionWith4Args<Connection, String, String, Boolean, Boolean> executeUpdateWithPropagateException;

  /**
   * Keeps resources handles by session id. Resources are of types @{@link HwcResource}
   * {@link #close()} method closes all of them and session as well.
   */
  private static final Map<String, Set<HwcResource>> RESOURCE_IDS_BY_SESSION_ID = new HashMap<>();

  private final String sessionId;
  private final AtomicReference<HwcSessionState> hwcSessionStateRef;

  static {
    try {
      // Initializing to avoid exception being logged, if the spark-shell is
      // terminated without any query through Llap path.
      LlapBaseInputFormat.close("non-existing-handle-id");
    } catch (IOException e) {
      LOG.debug("Error encountered while initializing LlapBaseInputFormat", e);
    }
  }

  private boolean usingDirectReaderMode() {
    return (HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V1) ||
            HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V2) ||
            HWConf.getHwcExecutionMode().equalsIgnoreCase(HWConf.HWC_EXECUTION_MODE_SPARK));
  }

  public HiveWarehouseSessionImpl(HiveWarehouseSessionState sessionState) {
    this.sessionState = sessionState;
    this.sessionId = UUID.randomUUID().toString();
    getConnector = () -> DefaultJDBCWrapper.getConnector(sessionState);
    getConnOutsidePool = () -> DefaultJDBCWrapper.getJDBCConnOutsideThePool(sessionState);
    executeStmt = (conn, database, sql) ->
            DefaultJDBCWrapper.executeStmt(conn, database, sql, 0);
    executeUpdate = (conn, database, sql) ->
        DefaultJDBCWrapper.executeUpdate(conn, database, sql);
    executeUpdateWithPropagateException = DefaultJDBCWrapper::executeUpdate;
    sessionState.session.listenerManager().register(new LlapQueryExecutionListener());
    sessionState.session.sparkContext().addSparkListener(new HwcSparkListener());
    hwcSessionStateRef = new AtomicReference<>(HwcSessionState.OPEN);

    if (usingDirectReaderMode()) {
      sessionState.session.conf().set("spark.kryo.registrator",
              "com.qubole.spark.hiveacid.util.HiveAcidKyroRegistrator");
    }
    LOG.info("Created a new HWC session: {}", sessionId);
  }

  private enum HwcSessionState {
    OPEN, CLOSED;
  }

  private void ensureSessionOpen() {
    Preconditions.checkState(HwcSessionState.OPEN.equals(hwcSessionStateRef.get()), SESSION_CLOSED_MSG);
  }

  public Dataset<Row> q(String sql) {
    ensureSessionOpen();
    return executeQuery(sql);
  }

  @Override
  public Dataset<Row> sql(String queryToFetchData) {

    if (HWCReadMode.resolveReadMode(HWCReadMode.LLAP)) {
      return executeQueryInternal(queryToFetchData, null);
    } else if (usingDirectReaderMode()) {
      //datasource v1 will handle both old and new direct-reader
      return executeQueryInternal(queryToFetchData, null);
    } else if (HWCReadMode.resolveReadMode(HWCReadMode.JDBC_CLIENT)) {
      return executeSimpleJdbc(queryToFetchData);
    }
    else if(HWCReadMode.resolveReadMode(HWCReadMode.STAGING_OUTPUT)){
      return executeQueryStagingOutput(queryToFetchData);
    }
    else {
      // jdbc cluster
      return executeJdbcInternal(queryToFetchData);
    }
  }

  public Dataset<Row> executeQuery(String sql) {
    return executeSmart(EXECUTE_QUERY_LLAP, sql, null);
  }

  public Dataset<Row> executeQuery(String sql, boolean useSplitsEqualToSparkCores) {
    return executeSmart(EXECUTE_QUERY_LLAP, sql, getCoresInSparkCluster());
  }

  public Dataset<Row> executeQuery(String sql, int numSplitsToDemand) {
    return executeSmart(EXECUTE_QUERY_LLAP, sql, numSplitsToDemand);
  }

  private Dataset<Row> executeQueryInternal(String sql, Integer numSplitsToDemand) {
    // If query is select query, then used HiveWarehouseDataSource or else use Simple JDBC in case of DDL operations.
    Preconditions.checkNotNull(sql,"The query cannot be null");
    if (isDataFetchQuery(sql)) {
      ensureSessionOpen();

      if(HWCReadMode.resolveReadMode(HWCReadMode.STAGING_OUTPUT)){
        return executeQueryStagingOutput(sql);
      }

      else if (usingDirectReaderMode()) {
        if (!HWConf.getSparkSqlExtension(sessionState).
                contains("com.qubole.spark.hiveacid.HiveAcidAutoConvertExtension")) {
          LOG.info("For spark execution, set " +
                  "spark.sql.extensions to com.qubole.spark.hiveacid.HiveAcidAutoConvertExtension");
          LOG.warn(" com.qubole.spark.hiveacid.HiveAcidAutoConvertExtension will be replaced by " +
                  "com.hortonworks.spark.sql.rule.Extensions." +
                  "Please switch to com.hortonworks.spark.sql.rule.Extensions.");
        }
        return session().sql(sql);
      }
      String mode = EXECUTE_QUERY_LLAP.name();
      DataFrameReader dfr = session().read().format(HIVE_WAREHOUSE_CONNECTOR_INTERNAL).option("query", sql)
              .option(HWC_SESSION_ID_KEY, sessionId).option(HWC_QUERY_MODE, mode);
      if (numSplitsToDemand != null) {
        dfr.option(JobUtil.SESSION_QUERIES_FOR_GET_NUM_SPLITS, setSplitPropertiesQuery(numSplitsToDemand));
      }
      return dfr.load();
    } else {
      return executeSimpleJdbc(sql);
    }
  }

  public static void addResourceIdToSession(String sessionId, HwcResource resourceId) {
    LOG.info("Adding resource: {} to current session: {}", resourceId, sessionId);
    synchronized (RESOURCE_IDS_BY_SESSION_ID) {
      RESOURCE_IDS_BY_SESSION_ID.putIfAbsent(sessionId, new HashSet<>());
      RESOURCE_IDS_BY_SESSION_ID.get(sessionId).add(resourceId);
    }
  }

  public static void closeAndRemoveResourceFromSession(String sessionId, HwcResource hwcResource) throws IOException {
    Set<HwcResource> hwcResources;
    boolean resourcePresent;
    synchronized (RESOURCE_IDS_BY_SESSION_ID) {
      hwcResources = RESOURCE_IDS_BY_SESSION_ID.get(sessionId);
      resourcePresent = hwcResources != null && hwcResources.remove(hwcResource);
    }
    if (resourcePresent) {
      LOG.info("Remove and close resource: {} from current session: {}", hwcResource, sessionId);
      hwcResource.close();
    }
  }

  private static void closeSessionResources(String sessionId) throws IOException {
    LOG.info("Closing all resources for current session: {}", sessionId);
    Set<HwcResource> hwcResources;
    synchronized (RESOURCE_IDS_BY_SESSION_ID) {
      hwcResources = RESOURCE_IDS_BY_SESSION_ID.remove(sessionId);
    }
    if (hwcResources != null && !hwcResources.isEmpty()) {
      for (HwcResource resource : hwcResources) {
        resource.close();
      }
    }
  }

  private String setSplitPropertiesQuery(int numSplitsToDemand) {
    return String.format("SET tez.grouping.split-count=%s", numSplitsToDemand);
  }

  private int getCoresInSparkCluster() {
    //this was giving good results(number of cores) when tested in standalone and yarn mode(without dynamicAllocation)
    return session().sparkContext().defaultParallelism();
  }

  public Dataset<Row> execute(String sql) {
    return executeSmart(EXECUTE_HIVE_JDBC, sql, null);
  }

  private Dataset<Row> executeSmart(ExecutionMethod current, String sql, Integer numSplitsToDemand) {
    ExecutionMethod resolved = resolveExecutionMethod(
        BooleanUtils.toBoolean(HWConf.SMART_EXECUTION.getString(sessionState)), current, sql);
    if (EXECUTE_HIVE_JDBC.equals(resolved)) {
      return executeSimpleJdbc(sql);
    }
    return executeQueryInternal(sql, numSplitsToDemand);
  }

  public String getJdbcUrlForExecutor() {
    RuntimeConfig conf = session().conf();
    if ((conf.contains(HWConf.HIVESERVER2_CREDENTIAL_ENABLED) && conf.get(HWConf.HIVESERVER2_CREDENTIAL_ENABLED).equals("true"))
            || conf.contains(HWConf.HIVESERVER2_JDBC_URL_PRINCIPAL)) {
      String master = conf.get(HWConf.SPARK_MASTER, "local");
      LOG.debug("Getting jdbc connection url for kerberized cluster with spark.master = {}", master);
      if (master.equals("yarn")) {
        StringBuilder jdbcURLBuilder = new StringBuilder();
        // In YARN mode for kerberized clusters, executors always run on yarn application. So, need
        // to use delegationToken to authenticate HS2 connection from it.
        jdbcURLBuilder.append(format("%s;auth=delegationToken", conf.get(HWConf.HIVESERVER2_JDBC_URL)));
        if (conf.contains(HWConf.HIVESERVER2_JDBC_URL_CONF_LIST) &&
                StringUtils.isNotBlank(conf.get(HWConf.HIVESERVER2_JDBC_URL_CONF_LIST))) {
          jdbcURLBuilder.append("?").append(conf.get(HWConf.HIVESERVER2_JDBC_URL_CONF_LIST));
        }
        return jdbcURLBuilder.toString();
      }
    }
    return HWConf.RESOLVED_HS2_URL.getString(sessionState);
  }

  private Dataset<Row> executeJdbcInternal(String sql) {
    // If query is select query, then used JDBCDataSource or else use Simple JDBC in case of DDL operations.
    Preconditions.checkNotNull(sql,"The query cannot be null");
    if (isDataFetchQuery(sql)) {
      ensureSessionOpen();
      String mode = EXECUTE_HIVE_JDBC.name();
      DataFrameReader dfr = session().read().format(HIVE_WAREHOUSE_CONNECTOR_INTERNAL).option("query", sql)
              .option(HWC_SESSION_ID_KEY, sessionId)
              .option(HWC_QUERY_MODE, mode)
              .option(QueryExecutionUtil.HIVE_JDBC_URL_FOR_EXECUTOR, getJdbcUrlForExecutor());
      return dfr.load();
    } else {
      return executeSimpleJdbc(sql);
    }
  }


  private Dataset<Row> executeQueryStagingOutput(final String sql) {
    LOG.info("execution mode :Staging Output FGAC");
    Preconditions.checkNotNull(sql,"The query cannot be null");

    String querySha = JobUtil.getQuerySha(sql);
    String tableName = "table_" + querySha;

    SparkSession sparkSession = SparkSession.active();

    STAGING_OUTPUT_DIR = JobUtil.HWCStagingOutputLocation(sparkSession);
    LOG.debug("staging dir" + STAGING_OUTPUT_DIR);

    String tableLocation = STAGING_OUTPUT_DIR + "/" + tableName;

    LOG.debug("Location=" + tableLocation);

    try {
      Configuration conf = new Configuration();
      FileSystem fs = FileSystem.get(conf);

      if (!fs.exists(new Path(tableLocation))) {

        FsPermission Permission = new FsPermission("1700");
        fs.mkdirs(new Path(tableLocation), Permission);
        LOG.debug("Creating staging table");

        String ctas = String.format("CREATE TEMPORARY  EXTERNAL TABLE %s STORED AS ORC LOCATION '%s'  AS SELECT * FROM (%s) as q",
                tableName, tableLocation, sql);
        LOG.info("query: " + ctas);
        STAGING_OUTPUT = true;

        try (Connection conn = getConnOutsidePool.get()) {
          executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState),
                  "set hive.exec.orc.delta.streaming.optimizations.enabled=true", true);
          executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState),
                  "set hive.stats.autogather=false", true);
          executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState),
                  "set tez.runtime.pipelined.sorter.lazy-allocate.memory=true", false);
          executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState), ctas, true);
          conn.close();
          LOG.debug("CTAS SUCCESS!");
        } catch (SQLException e) {
          // Ctas execution failed, so we need to delete the staging dir.
          JobUtil.deleteHWCStagingDir(tableLocation,STAGING_OUTPUT);
          STAGING_OUTPUT = false;
          throw new RuntimeException(e.getMessage(), e);
        }

      } else {
        LOG.info("Table already exists in staging space for query: {}", sql);
      }
    } catch (IOException e) {
      throw new RuntimeException(e.getMessage(), e);
    }

    return sparkSession.read().orc(tableLocation);

  }

  public static void closeAllSessionsResources() {
    LOG.debug(STAGING_OUTPUT + "delete staging: " + STAGING_OUTPUT_DIR );
      JobUtil.deleteHWCStagingDir(STAGING_OUTPUT_DIR, STAGING_OUTPUT);
  }

  public boolean executeUpdate(String sql) {
    return executeUpdate(sql, true);
  }

  @Override
  public boolean executeUpdate(String sql, boolean propagateException) {
    ensureSessionOpen();
    try (Connection conn = getConnector.get()) {
      return executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState), sql, propagateException);
    } catch (SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  private boolean executeUpdateInternal(String sql, Connection conn) {
    ensureSessionOpen();
    return executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState), sql, true);
  }

  public Dataset<Row> table(String table) {
    ensureSessionOpen();
    String mode = EXECUTE_HIVE_JDBC.name();
    if (BooleanUtils.toBoolean(HWConf.READ_VIA_LLAP.getString(sessionState))) {
      mode = EXECUTE_QUERY_LLAP.name();
    }
    DataFrameReader dfr = session().read().format(HIVE_WAREHOUSE_CONNECTOR_INTERNAL).option("table", table)
        .option(HWC_SESSION_ID_KEY, sessionId).option(HWC_QUERY_MODE, mode)
        .option(QueryExecutionUtil.HIVE_JDBC_URL_FOR_EXECUTOR, getJdbcUrlForExecutor());
    return dfr.load();
  }

  public SparkSession session() {
    ensureSessionOpen();
    return sessionState.session;
  }

  // Exposed for Python side.
  public HiveWarehouseSessionState sessionState() {
    return sessionState;
  }

  SparkConf conf() {
    return sessionState.session.sparkContext().getConf();
  }

  /* Catalog helpers */
  public void setDatabase(String name) {
    ensureSessionOpen();
    HWConf.DEFAULT_DB.setString(sessionState, name);
  }

  private Dataset<Row> executeSimpleJdbc(String sql) {
    ensureSessionOpen();
    try (Connection conn = getConnector.get()) {
      DriverResultSet drs = executeStmt.apply(conn, DEFAULT_DB.getString(sessionState), sql);
      return drs.asDataFrame(session());
    } catch (SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  public Dataset<Row> showDatabases() {
    ensureSessionOpen();
    return executeSimpleJdbc(HiveQlUtil.showDatabases());
  }

  public Dataset<Row> showTables() {
    ensureSessionOpen();
    return executeSimpleJdbc(HiveQlUtil.showTables(DEFAULT_DB.getString(sessionState)));
  }

  public Dataset<Row> describeTable(String table) {
    ensureSessionOpen();
    return executeSimpleJdbc(HiveQlUtil.describeTable(DEFAULT_DB.getString(sessionState), table));
  }

  public void dropDatabase(String database, boolean ifExists, boolean cascade) {
    executeUpdate(HiveQlUtil.dropDatabase(database, ifExists, cascade));
  }

  public void dropTable(String table, boolean ifExists, boolean purge) {
    ensureSessionOpen();
    try (Connection conn = getConnector.get()) {
      executeUpdateInternal(HiveQlUtil.useDatabase(DEFAULT_DB.getString(sessionState)), conn);
      String dropTable = HiveQlUtil.dropTable(table, ifExists, purge);
      executeUpdateInternal(dropTable, conn);
    } catch (SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  public void createDatabase(String database, boolean ifNotExists) {
    executeUpdate(HiveQlUtil.createDatabase(database, ifNotExists), true);
  }

  public CreateTableBuilder createTable(String tableName) {
    ensureSessionOpen();
    return new CreateTableBuilder(this, DEFAULT_DB.getString(sessionState), tableName, ORC, new HWCOptions(new HashMap<>()));
  }

  @Override
  public MergeBuilder mergeBuilder() {
    ensureSessionOpen();
    return new MergeBuilderImpl(this, DEFAULT_DB.getString(sessionState));
  }

  @Override
  public void cleanUpStreamingMeta(String queryCheckpointDir, String dbName, String tableName) {
    ensureSessionOpen();
    try {
      new StreamingMetaCleaner(sessionState, queryCheckpointDir, dbName, tableName).clean();
    } catch (IOException | SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  @Override
  public void commitTxn() {
    if (usingDirectReaderMode()) {
      HiveAcidTxnManagerObject.commitTxn(session());
    }
  }

  @Override
  public void close() {
    try {
      commitTxn();
    } catch (Exception e) {
      LOG.info("Failed to end hive-acid txn. The txn may be committed already.", e);
    }
    Preconditions.checkState(hwcSessionStateRef.compareAndSet(HwcSessionState.OPEN, HwcSessionState.CLOSED),
        SESSION_CLOSED_MSG);

    // close llap handles, broadcast variables
    try {
      closeSessionResources(sessionId);
    } catch (IOException e) {
      throw new RuntimeException("Error while closing resources attached to session: " + sessionId, e);
    } finally {
      // close all JDBC connections
      DefaultJDBCWrapper.closeHS2ConnectionPool();
    }
  }

  public String getSessionId() {
    return this.sessionId;
  }

  protected void finalize() {
    close();
  }
}

