package com.hortonworks.spark.sql.hive.llap.readers;

import com.google.common.base.Preconditions;
import com.hortonworks.hwc.plan.HwcPlannerStatistics;
import com.hortonworks.spark.sql.hive.llap.DefaultJDBCWrapper;
import com.hortonworks.spark.sql.hive.llap.FilterPushdown;
import com.hortonworks.spark.sql.hive.llap.common.DriverResultSet;
import com.hortonworks.spark.sql.hive.llap.common.HWConf;
import com.hortonworks.spark.sql.hive.llap.HiveWarehouseSessionImpl;
import com.hortonworks.spark.sql.hive.llap.common.SerializableLlapInputSplit;
import com.hortonworks.spark.sql.hive.llap.common.StatementType;
import com.hortonworks.spark.sql.hive.llap.common.CommonBroadcastInfo;
import com.hortonworks.spark.sql.hive.llap.common.HwcResource;
import com.hortonworks.spark.sql.hive.llap.readers.batch.HiveCountInputPartition;
import com.hortonworks.spark.sql.hive.llap.readers.batch.HiveWarehouseInputPartition;
import com.hortonworks.spark.sql.hive.llap.readers.row.HiveWarehouseRowInputPartition;
import com.hortonworks.spark.sql.hive.llap.util.JobUtil;
import com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil;
import com.hortonworks.spark.sql.hive.llap.util.SchemaUtil;
import org.apache.hadoop.hive.llap.LlapBaseInputFormat;
import org.apache.hadoop.hive.llap.LlapInputSplit;
import org.apache.hadoop.hive.llap.Schema;
import org.apache.hadoop.mapred.InputSplit;
import org.apache.hadoop.mapred.JobConf;
import org.apache.spark.SparkContext;
import org.apache.spark.broadcast.Broadcast;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.sql.catalyst.InternalRow;
import org.apache.spark.sql.internal.SQLConf;
import org.apache.spark.sql.sources.Filter;
import org.apache.spark.sql.sources.v2.reader.*;
import org.apache.spark.sql.types.StructType;
import org.apache.spark.sql.sources.v2.reader.DataSourceReader;
import org.apache.spark.sql.sources.v2.reader.SupportsScanColumnarBatch;
import org.apache.spark.sql.vectorized.ColumnarBatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.collection.Seq;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.OptionalLong;

import static com.hortonworks.spark.sql.hive.llap.FilterPushdown.buildWhereClause;
import static com.hortonworks.spark.sql.hive.llap.util.HiveQlUtil.*;
import static com.hortonworks.spark.sql.hive.llap.util.JobUtil.replaceSparkHiveDriver;
import static scala.collection.JavaConversions.asScalaBuffer;

/**
 * 1. Spark pulls the unpruned schema -> readSchema()
 * 2. Spark pulls factories, where factory/task are 1:1
 * -> if (enableBatchRead)
 * createBatchDataReaderFactories(..)
 * else
 * createDataReaderFactories(..)
 */
public class HiveWarehouseDataSourceReader
    implements DataSourceReader, SupportsPushDownRequiredColumns, SupportsScanColumnarBatch, SupportsPushDownFilters,
    SupportsReportStatistics {

  //The original schema
  protected StructType baseSchema = null;

  //The pruned schema
  protected StructType schema = null;

  //Pushed down filters
  //
  //"It's possible that there is no filters in the query and pushFilters(Filter[])
  // is never called, empty array should be returned for this case."
  Filter[] pushedFilters = new Filter[0];

  //SessionConfigSupport options
  protected Map<String, String> options;

  private static Logger LOG = LoggerFactory.getLogger(HiveWarehouseDataSourceReader.class);

  private final String sessionId;
  protected JobConf jobConf;
  protected CommonBroadcastInfo commonBroadcastInfo;

  // Enable ColumnarBatch reader by default.
  private boolean enableBatchRead;
  private HwcResource hwcResource;

  public HiveWarehouseDataSourceReader(Map<String, String> options) throws IOException {
    this.options = options;

    //this is a hack to prevent the following situation:
    //Spark(v 2.4.0) creates one instance of DataSourceReader to call readSchema() and then a new instance of DataSourceReader
    //to call pushFilters(), planBatchInputPartitions() etc. Since it uses different DataSourceReader instances,
    //and reads schema in former instance, schema remains null in the latter instance(which causes problems for other methods).
    //More discussion: http://apache-spark-user-list.1001560.n3.nabble.com/DataSourceV2-APIs-creating-multiple-instances-of-DataSourceReader-and-hence-not-preserving-the-state-tc33646.html
    //Also a null check on schema is already there in readSchema() to prevent initialization more than once just in case.
    readSchema();
    sessionId = getCurrentSessionId();
    enableBatchRead = shouldEnableBatchRead();
  }

  private boolean shouldEnableBatchRead() {
    // If the number of projected columns exceed the supported limit for batch reader, then use non-batch reader
    int columnsLimitForBatchDataReader
        = Integer.parseInt(HWConf.BATCH_DATAREADER_COLUMNS_LIMIT.getFromOptionsMap(this.options));
    Preconditions.checkState(columnsLimitForBatchDataReader > 0,
        HWConf.INVALID_BATCH_DATAREADER_COLUMNS_LIMIT_CONFIG_ERR_MSG);

    // Check if we need to enable batch read. Need to check based on input config.
    return (this.readSchema().length() <= columnsLimitForBatchDataReader);
  }

  //if(schema is empty) -> df.count()
  //else if(using table option) -> select *
  //else -> SELECT <COLUMNS> FROM (<RAW_SQL>) WHERE <FILTER_CLAUSE>
  String getQueryString(String[] requiredColumns, Filter[] filters) {
    String selectCols = "count(*)";
    if (requiredColumns.length > 0) {
      selectCols = projections(requiredColumns);
    }
    String baseQuery;
    if (getQueryType() == StatementType.FULL_TABLE_SCAN) {
      baseQuery = selectStar(options.get("table"));
    } else {
      baseQuery = options.get("query");
    }

    Seq<Filter> filterSeq = asScalaBuffer(Arrays.asList(filters)).seq();
    String whereClause = buildWhereClause(baseSchema, filterSeq);
    return selectProjectAliasFilter(selectCols, baseQuery, randomAlias(), whereClause);
  }


  private StatementType getQueryType() {
    return StatementType.fromOptions(options);
  }

  private String getCurrentSessionId() {
    String sessionId = options.get(HiveWarehouseSessionImpl.HWC_SESSION_ID_KEY);
    Preconditions.checkNotNull(sessionId,
        "session id cannot be null, forgot to initialize HiveWarehouseSession???");
    return sessionId;
  }

  protected StructType getTableSchema() throws Exception {
    replaceSparkHiveDriver();

    StatementType queryKey = getQueryType();
    String query;
    if (queryKey == StatementType.FULL_TABLE_SCAN) {
      String dbName = HWConf.DEFAULT_DB.getFromOptionsMap(options);
      SchemaUtil.TableRef tableRef = SchemaUtil.getDbTableNames(dbName, options.get("table"));
      query = selectStar(tableRef.databaseName, tableRef.tableName);
    } else {
      query = options.get("query");
    }
    LlapBaseInputFormat llapInputFormat = null;
    try {
      JobConf conf = JobUtil.createJobConf(options, query);
      llapInputFormat = new LlapBaseInputFormat(false, Long.MAX_VALUE);
      InputSplit[] splits = llapInputFormat.getSplits(conf, 0);
      LlapInputSplit schemaSplit = (LlapInputSplit) splits[0];
      Schema schema = schemaSplit.getSchema();
      return SchemaUtil.convertSchema(schema);
    } finally {
      if (llapInputFormat != null) {
        close();
      }
    }
  }

  @Override
  public StructType readSchema() {
    try {
      if (schema == null) {
        this.schema = getTableSchema();
        this.baseSchema = this.schema;
      }
      return schema;
    } catch (Exception e) {
      LOG.error("Unable to read table schema");
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  //"returns unsupported filters."
  @Override
  public Filter[] pushFilters(Filter[] filters) {
    pushedFilters = Arrays.stream(filters).
        filter((filter) -> FilterPushdown.buildFilterExpression(baseSchema, filter).isDefined()).
        toArray(Filter[]::new);

    return Arrays.stream(filters).
        filter((filter) -> !FilterPushdown.buildFilterExpression(baseSchema, filter).isDefined()).
        toArray(Filter[]::new);
  }

  @Override
  public Filter[] pushedFilters() {
    return pushedFilters;
  }

  @Override
  public void pruneColumns(StructType requiredSchema) {
    this.schema = requiredSchema;
  }


  /* Method used by Spark executor if enableBatchRead = false */
  @Override
  public List<InputPartition<InternalRow>> planInputPartitions() {
    LOG.info("Creating non-batch data reader factories.");
    try {
      assert (this.schema.length() > 0);
      String queryString = getQueryString(SchemaUtil.columnNames(schema), pushedFilters);
      return getSplitsInputPartitions(queryString);
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /* Method used by Spark executor if enableBatchRead = true */
  @Override
  public List<InputPartition<ColumnarBatch>> planBatchInputPartitions() {
    boolean countStar = this.schema.length() == 0;
    String queryString = getQueryString(SchemaUtil.columnNames(schema), pushedFilters);
    List<InputPartition<ColumnarBatch>> factories = new ArrayList<>();
    if (countStar) {
      LOG.info("Executing count with query: {}", queryString);
      factories.addAll(getCountStarInputPartitions(queryString));
    } else {
      factories.addAll(getSplitsInputPartitions(queryString));
    }
    return factories;
  }

  private List<InputPartition<ColumnarBatch>> getCountStarInputPartitions(String query) {
    List<InputPartition<ColumnarBatch>> tasks = new ArrayList<>(100);
    long count = getCount(query);
    String numTasksString = HWConf.COUNT_TASKS.getFromOptionsMap(options);
    int numTasks = Integer.parseInt(numTasksString);
    long numPerTask = count / (numTasks - 1);
    long numLastTask = count % (numTasks - 1);
    for (int i = 0; i < (numTasks - 1); i++) {
      tasks.add(new HiveCountInputPartition(numPerTask));
    }
    tasks.add(new HiveCountInputPartition(numLastTask));
    return tasks;
  }


  protected long getCount(String query) {
    try (Connection conn = QueryExecutionUtil.getConnection(options)) {
      DriverResultSet rs = DefaultJDBCWrapper.executeStmt(conn, HWConf.DEFAULT_DB.getFromOptionsMap(options), query,
          Long.parseLong(HWConf.MAX_EXEC_RESULTS.getFromOptionsMap(options)));
      long count = rs.getData().get(0).getLong(0);
      return count;
    } catch (SQLException e) {
      LOG.error("Failed to connect to HS2", e);
      throw new RuntimeException(e.getMessage(), e);
    }
  }


  @Override
  public boolean enableBatchRead() {
    return enableBatchRead;
  }

  // TODO: Giving up generics - we need InputPartition<ColumnarBatch> and InputPartition<Row>
  protected List getSplitsInputPartitions(String query) {
    List tasks = new ArrayList<>();
    try {
      JobConf jobConf = JobUtil.createJobConf(options, query);
      LlapBaseInputFormat llapInputFormat = new LlapBaseInputFormat(false, Long.MAX_VALUE);
      LOG.info("Additional props for generating splits: {}", options.get(JobUtil.SESSION_QUERIES_FOR_GET_NUM_SPLITS));
      //numSplits arg not currently supported, use 1 as dummy arg
      InputSplit[] splits = llapInputFormat.getSplits(jobConf, 1);
      LOG.info("Number of splits generated: {}", splits.length);

      if (splits.length > 2) {
        commonBroadcastInfo = prepareCommonBroadcastInfo(splits);
        LOG.info("Serializing {} actual splits to send to executors", (splits.length - 2));
        byte[] serializedJobConf = JobUtil.serializeJobConf(jobConf);

        long start = System.currentTimeMillis();
        for (int i = 2; i < splits.length; i++) {
          LlapInputSplit actualSplit = (LlapInputSplit) splits[i];
          tasks.add(getInputPartition(actualSplit, serializedJobConf, getArrowAllocatorMax(), commonBroadcastInfo));
        }
        long end = System.currentTimeMillis();
        LOG.info("Serialized {} actual splits in {} millis", (splits.length - 2), (end - start));
      } else {
        LOG.warn("No actual splits generated for query: {}", query);
      }
    } catch (IOException e) {
      LOG.error("Unable to submit query to HS2");
      throw new RuntimeException(e.getMessage(), e);
    } finally {
      // add handle id for HiveWarehouseSessionImpl#close()
      hwcResource = new HwcResource(options.get(JobUtil.LLAP_HANDLE_ID), commonBroadcastInfo);
      HiveWarehouseSessionImpl.addResourceIdToSession(sessionId, hwcResource);
    }
    return tasks;
  }

  // TODO: Giving up generics - we need InputPartition<ColumnarBatch> and InputPartition<Row>
  protected InputPartition getInputPartition(InputSplit split, byte[] serializedJobConf,
                                                long arrowAllocatorMax, CommonBroadcastInfo commonBroadcastInfo) {
    if (enableBatchRead) {
      return (new HiveWarehouseInputPartition(split, serializedJobConf,
          arrowAllocatorMax, commonBroadcastInfo));
    } else {
      return (new HiveWarehouseRowInputPartition(split, serializedJobConf,
          arrowAllocatorMax, commonBroadcastInfo));
    }
  }

  protected CommonBroadcastInfo prepareCommonBroadcastInfo(InputSplit[] splits) {
    SparkContext sparkContext = SparkSession.getActiveSession().get().sparkContext();

    // populate actual splits with schema and planBytes[]
    LlapInputSplit schemaSplit = (LlapInputSplit) splits[0];
    LlapInputSplit planSplit = (LlapInputSplit) splits[1];

    // Don't broadcast if splits have already been broadcast for some previous execution
    if (commonBroadcastInfo != null
        && commonBroadcastInfo.getSchemaSplit().isValid()
        && commonBroadcastInfo.getPlanSplit().isValid()
        && planSplit.equals(commonBroadcastInfo.getPlanSplit().getValue().getLlapInputSplit())
        && schemaSplit.equals(commonBroadcastInfo.getSchemaSplit().getValue().getLlapInputSplit())) {
      return commonBroadcastInfo;
    }

    Broadcast<SerializableLlapInputSplit> planSplitBroadcast =
        broadcastSplit(sparkContext, planSplit);

    Broadcast<SerializableLlapInputSplit> schemaSplitBroadcast =
        broadcastSplit(sparkContext, schemaSplit);

    CommonBroadcastInfo commonBroadcastInfo = new CommonBroadcastInfo();
    commonBroadcastInfo.setSchemaSplit(schemaSplitBroadcast);
    commonBroadcastInfo.setPlanSplit(planSplitBroadcast);
    return commonBroadcastInfo;
  }

  private Broadcast<SerializableLlapInputSplit> broadcastSplit(SparkContext sparkContext, LlapInputSplit planSplit) {
    return sparkContext.broadcast(new SerializableLlapInputSplit(planSplit),
        SchemaUtil.classTag(SerializableLlapInputSplit.class));
  }

  // spark uses these stats to generate better plan for queries
  // for now we only support for StatementType.FULL_TABLE_SCAN as we can determine stats for them.
  // If reader is invoked via translation rule (Extensions.scala) then we automatically get StatementType.FULL_TABLE_SCAN
  // as we have one-to-one mapping of table to reader.
  @Override
  public Statistics estimateStatistics() {
    SQLConf conf = SparkSession.getActiveSession().get().sqlContext().conf();
    // initialize with default stats
    HwcPlannerStatistics statistics = new HwcPlannerStatistics(OptionalLong.of(conf.defaultSizeInBytes()),
        OptionalLong.empty());

    try (Connection conn = QueryExecutionUtil.getConnection(options)) {
      if (StatementType.FULL_TABLE_SCAN.equals(getQueryType())) {
        String dbName = HWConf.DEFAULT_DB.getFromOptionsMap(options);
        SchemaUtil.TableRef tableRef = SchemaUtil.getDbTableNames(dbName, options.get("table"));
        LOG.debug("Found full table scan for table {}.{}, estimating stastics",
            tableRef.databaseName, tableRef.tableName);
        statistics = JobUtil.getStatisticsForTable(tableRef.databaseName, tableRef.tableName, conn);
        LOG.debug("Estimated statistics {} for table {}.{}", statistics, tableRef.databaseName, tableRef.tableName);
      }
    } catch (Exception e) {
      LOG.error("Unable to get table properties, setting to defaults " + statistics
          + " Error message - " + e.getMessage(), e);
    }
    return statistics;
  }


  private long getArrowAllocatorMax() {
    String arrowAllocatorMaxString = HWConf.ARROW_ALLOCATOR_MAX.getFromOptionsMap(options);
    long arrowAllocatorMax = (Long) HWConf.ARROW_ALLOCATOR_MAX.getDefaultValue();
    if (arrowAllocatorMaxString != null) {
      arrowAllocatorMax = Long.parseLong(arrowAllocatorMaxString);
    }
    LOG.debug("Ceiling for Arrow direct buffers {}", arrowAllocatorMax);
    return arrowAllocatorMax;
  }

  public void close() {
    try {
      HiveWarehouseSessionImpl.closeAndRemoveResourceFromSession(sessionId, hwcResource);
    } catch (IOException ioe) {
      throw new RuntimeException(ioe.getMessage(), ioe);
    }
  }

}
