package com.hortonworks.spark.sql.hive.llap.readers.batch;

import com.hortonworks.spark.sql.hive.llap.common.CommonBroadcastInfo;
import com.hortonworks.spark.sql.hive.llap.util.JobUtil;
import org.apache.arrow.vector.FieldVector;
import org.apache.arrow.vector.TimeStampMicroTZVector;
import org.apache.arrow.vector.types.pojo.ArrowType;
import org.apache.hadoop.hive.llap.LlapBaseInputFormat;
import org.apache.hadoop.hive.llap.LlapInputSplit;
import org.apache.hadoop.hive.ql.io.arrow.ArrowWrapperWritable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.RecordReader;
import org.apache.hadoop.mapreduce.MRJobConfig;
import org.apache.spark.sql.sources.v2.reader.InputPartitionReader;
import org.apache.spark.sql.vectorized.ArrowColumnVector;
import org.apache.spark.sql.vectorized.ColumnVector;
import org.apache.spark.sql.vectorized.ColumnarBatch;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.TimeZone;

import org.apache.hadoop.mapreduce.TaskAttemptID;
import org.apache.hadoop.mapreduce.TaskID;
import org.apache.hadoop.mapreduce.JobID;
import org.apache.hadoop.yarn.api.records.ApplicationId;
import org.apache.hadoop.mapreduce.MRJobConfig;
import org.apache.hadoop.hive.llap.SubmitWorkInfo;
import org.apache.hadoop.mapreduce.TaskType;
import org.apache.arrow.memory.BufferAllocator;
import org.apache.hadoop.hive.ql.io.arrow.RootAllocatorFactory;

import static com.hortonworks.spark.sql.hive.llap.util.JobUtil.UTC_TIME_ZONE;

public class HiveWarehouseInputPartitionReader implements InputPartitionReader<ColumnarBatch> {


  private RecordReader<?, ArrowWrapperWritable> reader;
  private ArrowWrapperWritable wrapperWritable = new ArrowWrapperWritable();
  //Reuse single instance of ColumnarBatch and ColumnVector[]
  //See org.apache.spark.sql.vectorized.ColumnarBatch
  //"Instance of it is meant to be reused during the entire data loading process."
  private ColumnarBatch columnarBatch;
  private ColumnVector[] columnVectors;

  private String attemptId;
  private boolean inputExhausted = false;
  private final TimeZone sparkSessionTimezone;
  private final boolean hasSameTZRules;

  public HiveWarehouseInputPartitionReader(LlapInputSplit split, JobConf conf, long arrowAllocatorMax,
                                           CommonBroadcastInfo commonBroadcastInfo) throws Exception {

    commonBroadcastInfo.getPlanSplit().assertValid();
    split.setPlanBytes(commonBroadcastInfo.getPlanSplit().getValue().getLlapInputSplit().getPlanBytes());
    commonBroadcastInfo.getSchemaSplit().assertValid();
    split.setSchema(commonBroadcastInfo.getSchemaSplit().getValue().getLlapInputSplit().getSchema());

    // Set TASK_ATTEMPT_ID to submit to LlapOutputFormatService
    this.attemptId = JobUtil.getTaskAttemptID(split).toString();
    conf.set(MRJobConfig.TASK_ATTEMPT_ID, this.attemptId);
    this.reader = getRecordReader(split, conf, arrowAllocatorMax);

    // Driver side code always populates spark.sql.session.timeZone in conf
    this.sparkSessionTimezone = JobUtil.getSparkSessionTimezoneAtExecutor(conf);
    this.hasSameTZRules = UTC_TIME_ZONE.hasSameRules(this.sparkSessionTimezone);
  }

  // Used only for test.
  protected HiveWarehouseInputPartitionReader(LlapInputSplit split, JobConf conf, long arrowAllocatorMax)
          throws Exception {
    this.reader = getRecordReader(split, conf, arrowAllocatorMax);
    this.sparkSessionTimezone = TimeZone.getDefault();
    this.hasSameTZRules = UTC_TIME_ZONE.hasSameRules(this.sparkSessionTimezone);
  }

  protected RecordReader<?, ArrowWrapperWritable> getRecordReader(
          LlapInputSplit split, JobConf conf, long arrowAllocatorMax) throws IOException {
    return JobUtil.getLlapArrowBatchRecordReader(split, conf, arrowAllocatorMax, attemptId);
  }

  @Override
  public boolean next() throws IOException {
    // don't call reader.next() if the input is already exhausted or else it will block.
    // see https://hortonworks.jira.com/browse/BUG-120380 for more on it.
    if (inputExhausted) {
      return false;
    }
    boolean hasNextBatch = reader.next(null, wrapperWritable);
    inputExhausted = !hasNextBatch;
    return hasNextBatch;
  }

  @Override
  public ColumnarBatch get() {
    //Spark asks you to convert one column at a time so that different
    //column types can be handled differently.
    //NumOfCols << NumOfRows so this is negligible
    List<FieldVector> fieldVectors = wrapperWritable.getVectorSchemaRoot().getFieldVectors();
    if(columnVectors == null) {
      //Lazy create ColumnarBatch/ColumnVector[] instance
      columnVectors = new ColumnVector[fieldVectors.size()];
      columnarBatch = new ColumnarBatch(columnVectors);
    }
    Iterator<FieldVector> iterator = fieldVectors.iterator();
    int rowCount = -1;
    for (int i = 0; i < columnVectors.length; i++) {
      FieldVector fieldVector = iterator.next();
      if (ArrowType.Timestamp.TYPE_TYPE == fieldVector.getField().getType().getTypeID() && !hasSameTZRules) {
        // we need to adjust timestamp millis.
        // LLAP always gives us millis corresponding to UTC.
        // Here we shift millis according to jvm timezone. This is to follow same semantics as ORC reader.
        // And this makes output of HWC consistent with spark.sql() for an external ORC table.
        // Later if user chooses to modify "spark.sql.session.timeZone", spark modifies timestamp accordingly
        adjustTimestampMillis((TimeStampMicroTZVector) fieldVector);
      }

      columnVectors[i] = new ArrowColumnVector(fieldVector);
      if (rowCount == -1) {
        //All column vectors have same length so we can get rowCount from any column
        rowCount = fieldVector.getValueCount();
      }
    }
    columnarBatch.setNumRows(rowCount);
    return columnarBatch;
  }

  @Override
  public void close() throws IOException {
    //close() single ColumnarBatch instance
    if(this.columnarBatch != null) {
      this.columnarBatch.close();
    }
    //reader.close() will throw exception unless all arrow buffers have been released
    //See org.apache.hadoop.hive.llap.close()
    //See org.apache.arrow.memory.BaseAllocator.close()
    if(this.reader != null) {
      this.reader.close();
    }
  }

  private void adjustTimestampMillis(TimeStampMicroTZVector vector) {
    final int length = vector.getValueCapacity();
    for (int i = 0; i < length; i++) {
      if (vector.isSet(i) == 1) {
        final long microsInUTC = vector.get(i);
        final long microsInCurrentTimezone = JobUtil.convertMicros(microsInUTC, UTC_TIME_ZONE, sparkSessionTimezone);
        vector.set(i, microsInCurrentTimezone);
      }
    }
  }
}

